"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sarus_data_spec.protobuf.statistics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Size(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    DATASET_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    STATISTICS_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    dataset: typing.Text = ...
    """uuid"""

    name: typing.Text = ...
    @property
    def statistics(self) -> sarus_data_spec.protobuf.statistics_pb2.Statistics: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        dataset : typing.Text = ...,
        name : typing.Text = ...,
        statistics : typing.Optional[sarus_data_spec.protobuf.statistics_pb2.Statistics] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataset",b"dataset","name",b"name","properties",b"properties","statistics",b"statistics","uuid",b"uuid"]) -> None: ...
global___Size = Size
