# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/type.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import predicate_pb2 as sarus__data__spec_dot_protobuf_dot_predicate__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#sarus_data_spec/protobuf/type.proto\x12\x0fsarus_data_spec\x1a(sarus_data_spec/protobuf/predicate.proto\x1a#sarus_data_spec/protobuf/path.proto\"\xfa\x18\n\x04Type\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04null\x18\x03 \x01(\x0b\x32\x1a.sarus_data_spec.Type.NullH\x00\x12*\n\x04unit\x18\x04 \x01(\x0b\x32\x1a.sarus_data_spec.Type.UnitH\x00\x12\x30\n\x07\x62oolean\x18\x05 \x01(\x0b\x32\x1d.sarus_data_spec.Type.BooleanH\x00\x12\x30\n\x07integer\x18\x06 \x01(\x0b\x32\x1d.sarus_data_spec.Type.IntegerH\x00\x12*\n\x04\x65num\x18\x07 \x01(\x0b\x32\x1a.sarus_data_spec.Type.EnumH\x00\x12,\n\x05\x66loat\x18\x08 \x01(\x0b\x32\x1b.sarus_data_spec.Type.FloatH\x00\x12*\n\x04text\x18\t \x01(\x0b\x32\x1a.sarus_data_spec.Type.TextH\x00\x12,\n\x05\x62ytes\x18\n \x01(\x0b\x32\x1b.sarus_data_spec.Type.BytesH\x00\x12.\n\x06struct\x18\x0b \x01(\x0b\x32\x1c.sarus_data_spec.Type.StructH\x00\x12,\n\x05union\x18\x0c \x01(\x0b\x32\x1b.sarus_data_spec.Type.UnionH\x00\x12\x32\n\x08optional\x18\r \x01(\x0b\x32\x1e.sarus_data_spec.Type.OptionalH\x00\x12*\n\x04list\x18\x0e \x01(\x0b\x32\x1a.sarus_data_spec.Type.ListH\x00\x12,\n\x05\x61rray\x18\x0f \x01(\x0b\x32\x1b.sarus_data_spec.Type.ArrayH\x00\x12\x32\n\x08\x64\x61tetime\x18\x10 \x01(\x0b\x32\x1e.sarus_data_spec.Type.DatetimeH\x00\x12\x38\n\x0b\x63onstrained\x18\x11 \x01(\x0b\x32!.sarus_data_spec.Type.ConstrainedH\x00\x12\x36\n\nhypothesis\x18\x12 \x01(\x0b\x32 .sarus_data_spec.Type.HypothesisH\x00\x12&\n\x02id\x18\x13 \x01(\x0b\x32\x18.sarus_data_spec.Type.IdH\x00\x12*\n\x04\x64\x61te\x18\x14 \x01(\x0b\x32\x1a.sarus_data_spec.Type.DateH\x00\x12*\n\x04time\x18\x15 \x01(\x0b\x32\x1a.sarus_data_spec.Type.TimeH\x00\x12\x32\n\x08\x64uration\x18\x16 \x01(\x0b\x32\x1e.sarus_data_spec.Type.DurationH\x00\x12\x39\n\nproperties\x18\x02 \x03(\x0b\x32%.sarus_data_spec.Type.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x06\n\x04Null\x1a\x06\n\x04Unit\x1a\t\n\x07\x42oolean\x1a\xd0\x01\n\x07Integer\x12\x30\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\".sarus_data_spec.Type.Integer.Base\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x03\"`\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06UINT64\x10\x04\x12\n\n\x06UINT32\x10\x05\x12\n\n\x06UINT16\x10\x06\x12\t\n\x05UINT8\x10\x07\x1a\xde\x01\n\x04\x45num\x12-\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Enum.Base\x12\x0f\n\x07ordered\x18\x02 \x01(\x08\x12\x39\n\x0bname_values\x18\x03 \x03(\x0b\x32$.sarus_data_spec.Type.Enum.NameValue\x1a(\n\tNameValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03\"1\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x1a\x99\x01\n\x05\x46loat\x12.\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32 .sarus_data_spec.Type.Float.Base\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x01\"-\n\x04\x42\x61se\x12\x0b\n\x07\x46LOAT64\x10\x00\x12\x0b\n\x07\x46LOAT32\x10\x01\x12\x0b\n\x07\x46LOAT16\x10\x02\x1a\x31\n\x04Text\x12\x10\n\x08\x65ncoding\x18\x01 \x01(\t\x12\x17\n\x0fpossible_values\x18\x02 \x03(\t\x1a\x07\n\x05\x42ytes\x1ax\n\x06Struct\x12\x32\n\x06\x66ields\x18\x01 \x03(\x0b\x32\".sarus_data_spec.Type.Struct.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1av\n\x05Union\x12\x31\n\x06\x66ields\x18\x01 \x03(\x0b\x32!.sarus_data_spec.Type.Union.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x08Optional\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a=\n\x04List\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\x10\n\x08max_size\x18\x02 \x01(\x03\x1a;\n\x05\x41rray\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05shape\x18\x02 \x03(\x03\x1a\xb0\x01\n\x08\x44\x61tetime\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12\x31\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32#.sarus_data_spec.Type.Datetime.Base\".\n\x04\x42\x61se\x12\x0c\n\x08INT64_NS\x10\x00\x12\x0c\n\x08INT64_MS\x10\x01\x12\n\n\x06STRING\x10\x02\x1a\x97\x01\n\x04\x44\x61te\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12-\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Date.Base\"\x1d\n\x04\x42\x61se\x12\t\n\x05INT32\x10\x00\x12\n\n\x06STRING\x10\x01\x1a\xb6\x01\n\x04Time\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12-\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Time.Base\"<\n\x04\x42\x61se\x12\x0c\n\x08INT64_NS\x10\x00\x12\x0c\n\x08INT32_MS\x10\x01\x12\n\n\x06STRING\x10\x02\x12\x0c\n\x08INT64_US\x10\x03\x1aK\n\x08\x44uration\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x03\x1a\x62\n\x0b\x43onstrained\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12.\n\nconstraint\x18\x02 \x01(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x82\x01\n\nHypothesis\x12\x36\n\x05types\x18\x02 \x03(\x0b\x32\'.sarus_data_spec.Type.Hypothesis.Scored\x1a<\n\x06Scored\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05score\x18\x02 \x01(\x01\x1a\xb5\x01\n\x02Id\x12+\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1d.sarus_data_spec.Type.Id.Base\x12\x0e\n\x06unique\x18\x02 \x01(\x08\x12(\n\treference\x18\x03 \x01(\x0b\x32\x15.sarus_data_spec.Path\"H\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06STRING\x10\x04\x12\t\n\x05\x42YTES\x10\x05\x42\x06\n\x04typeb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.type_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TYPE_PROPERTIESENTRY._options = None
  _TYPE_PROPERTIESENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=136
  _TYPE._serialized_end=3330
  _TYPE_PROPERTIESENTRY._serialized_start=1167
  _TYPE_PROPERTIESENTRY._serialized_end=1216
  _TYPE_NULL._serialized_start=1218
  _TYPE_NULL._serialized_end=1224
  _TYPE_UNIT._serialized_start=1226
  _TYPE_UNIT._serialized_end=1232
  _TYPE_BOOLEAN._serialized_start=1234
  _TYPE_BOOLEAN._serialized_end=1243
  _TYPE_INTEGER._serialized_start=1246
  _TYPE_INTEGER._serialized_end=1454
  _TYPE_INTEGER_BASE._serialized_start=1358
  _TYPE_INTEGER_BASE._serialized_end=1454
  _TYPE_ENUM._serialized_start=1457
  _TYPE_ENUM._serialized_end=1679
  _TYPE_ENUM_NAMEVALUE._serialized_start=1588
  _TYPE_ENUM_NAMEVALUE._serialized_end=1628
  _TYPE_ENUM_BASE._serialized_start=1358
  _TYPE_ENUM_BASE._serialized_end=1407
  _TYPE_FLOAT._serialized_start=1682
  _TYPE_FLOAT._serialized_end=1835
  _TYPE_FLOAT_BASE._serialized_start=1790
  _TYPE_FLOAT_BASE._serialized_end=1835
  _TYPE_TEXT._serialized_start=1837
  _TYPE_TEXT._serialized_end=1886
  _TYPE_BYTES._serialized_start=1888
  _TYPE_BYTES._serialized_end=1895
  _TYPE_STRUCT._serialized_start=1897
  _TYPE_STRUCT._serialized_end=2017
  _TYPE_STRUCT_FIELD._serialized_start=1959
  _TYPE_STRUCT_FIELD._serialized_end=2017
  _TYPE_UNION._serialized_start=2019
  _TYPE_UNION._serialized_end=2137
  _TYPE_UNION_FIELD._serialized_start=1959
  _TYPE_UNION_FIELD._serialized_end=2017
  _TYPE_OPTIONAL._serialized_start=2139
  _TYPE_OPTIONAL._serialized_end=2186
  _TYPE_LIST._serialized_start=2188
  _TYPE_LIST._serialized_end=2249
  _TYPE_ARRAY._serialized_start=2251
  _TYPE_ARRAY._serialized_end=2310
  _TYPE_DATETIME._serialized_start=2313
  _TYPE_DATETIME._serialized_end=2489
  _TYPE_DATETIME_BASE._serialized_start=2443
  _TYPE_DATETIME_BASE._serialized_end=2489
  _TYPE_DATE._serialized_start=2492
  _TYPE_DATE._serialized_end=2643
  _TYPE_DATE_BASE._serialized_start=2614
  _TYPE_DATE_BASE._serialized_end=2643
  _TYPE_TIME._serialized_start=2646
  _TYPE_TIME._serialized_end=2828
  _TYPE_TIME_BASE._serialized_start=2768
  _TYPE_TIME_BASE._serialized_end=2828
  _TYPE_DURATION._serialized_start=2830
  _TYPE_DURATION._serialized_end=2905
  _TYPE_CONSTRAINED._serialized_start=2907
  _TYPE_CONSTRAINED._serialized_end=3005
  _TYPE_HYPOTHESIS._serialized_start=3008
  _TYPE_HYPOTHESIS._serialized_end=3138
  _TYPE_HYPOTHESIS_SCORED._serialized_start=3078
  _TYPE_HYPOTHESIS_SCORED._serialized_end=3138
  _TYPE_ID._serialized_start=3141
  _TYPE_ID._serialized_end=3322
  _TYPE_ID_BASE._serialized_start=3250
  _TYPE_ID_BASE._serialized_end=3322
# @@protoc_insertion_point(module_scope)
