syntax = "proto3";
package sarus_data_spec;


enum ConstraintKind {
  SYNTHETIC = 0;
  DP = 1;
  PUBLIC = 2;
  PEP = 3;
  MOCK = 4;
  PEP_FOR_REWRITING = 5;
}

message VariantConstraint {
  string uuid = 1;
  string dataspec = 2;
  ConstraintKind constraint_kind = 3;
  repeated string required_context = 4;
  map<string, string> properties = 5;
}
