# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/transform.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import scalar_pb2 as sarus__data__spec_dot_protobuf_dot_scalar__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/transform.proto\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\x1a%sarus_data_spec/protobuf/scalar.proto\"\x8d#\n\tTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12\x1d\n\x04spec\x18\x04 \x01(\x0b\x32\x0f.Transform.Spec\x12.\n\nproperties\x18\x05 \x03(\x0b\x32\x1a.Transform.PropertiesEntry\x12\x12\n\ninversible\x18\x06 \x01(\x08\x12\x19\n\x11schema_preserving\x18\x07 \x01(\x08\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xa3\x0c\n\x04Spec\x12\'\n\x08identity\x18\x01 \x01(\x0b\x32\x13.Transform.IdentityH\x00\x12\'\n\x08variable\x18\x02 \x01(\x0b\x32\x13.Transform.VariableH\x00\x12\'\n\x08\x63omposed\x18\x03 \x01(\x0b\x32\x13.Transform.ComposedH\x00\x12%\n\x07project\x18\x04 \x01(\x0b\x32\x12.Transform.ProjectH\x00\x12#\n\x06\x66ilter\x18\x05 \x01(\x0b\x32\x11.Transform.FilterH\x00\x12%\n\x07shuffle\x18\x06 \x01(\x0b\x32\x12.Transform.ShuffleH\x00\x12\x1f\n\x04join\x18\x07 \x01(\x0b\x32\x0f.Transform.JoinH\x00\x12\x1f\n\x04\x63\x61st\x18\x08 \x01(\x0b\x32\x0f.Transform.CastH\x00\x12#\n\x06sample\x18\t \x01(\x0b\x32\x11.Transform.SampleH\x00\x12\x30\n\ruser_settings\x18\n \x01(\x0b\x32\x17.Transform.UserSettingsH\x00\x12-\n\x0fprotect_dataset\x18\x0b \x01(\x0b\x32\x12.Transform.ProtectH\x00\x12\'\n\x08\x65xternal\x18\x0c \x01(\x0b\x32\x13.Transform.ExternalH\x00\x12)\n\tsynthetic\x18\r \x01(\x0b\x32\x14.Transform.SyntheticH\x00\x12)\n\ttranscode\x18\x0e \x01(\x0b\x32\x14.Transform.TranscodeH\x00\x12\x38\n\x11inverse_transcode\x18\x0f \x01(\x0b\x32\x1b.Transform.InverseTranscodeH\x00\x12&\n\x08get_item\x18\x10 \x01(\x0b\x32\x12.Transform.GetItemH\x00\x12\x34\n\x0fprotected_paths\x18\x11 \x01(\x0b\x32\x19.Transform.ProtectedPathsH\x00\x12\x43\n\x17\x61utomatic_user_settings\x18\x12 \x01(\x0b\x32 .Transform.AutomaticUserSettingsH\x00\x12.\n\x0cpublic_paths\x18\x13 \x01(\x0b\x32\x16.Transform.PublicPathsH\x00\x12\x30\n\rassign_budget\x18\x14 \x01(\x0b\x32\x17.Transform.AssignBudgetH\x00\x12\x36\n\x10\x61utomatic_budget\x18\x15 \x01(\x0b\x32\x1a.Transform.AutomaticBudgetH\x00\x12\x37\n\x10\x61ttribute_budget\x18\x16 \x01(\x0b\x32\x1b.Transform.AttributesBudgetH\x00\x12(\n\tsd_budget\x18\x17 \x01(\x0b\x32\x13.Transform.SDBudgetH\x00\x12,\n\x0b\x64\x65rive_seed\x18\x18 \x01(\x0b\x32\x15.Transform.DeriveSeedH\x00\x12+\n\x0bgroup_by_pe\x18\x19 \x01(\x0b\x32\x14.Transform.GroupByPEH\x00\x12\x34\n\x0fsampling_ratios\x18\x1a \x01(\x0b\x32\x19.Transform.SamplingRatiosH\x00\x12*\n\nselect_sql\x18\x1b \x01(\x0b\x32\x14.Transform.SelectSqlH\x00\x12%\n\x07\x65xtract\x18\x1c \x01(\x0b\x32\x12.Transform.ExtractH\x00\x12\x38\n\x11relationship_spec\x18\x1d \x01(\x0b\x32\x1b.Transform.RelationshipSpecH\x00\x12@\n\x15\x64ifferentiated_sample\x18\x1e \x01(\x0b\x32\x1f.Transform.DifferentiatedSampleH\x00\x12;\n\x13validated_user_type\x18\x1f \x01(\x0b\x32\x1c.Transform.ValidatedUserTypeH\x00\x12/\n\rdp_select_sql\x18  \x01(\x0b\x32\x16.Transform.DPSelectSqlH\x00\x12\x36\n\x10\x65rror_estimation\x18! \x01(\x0b\x32\x1a.Transform.ErrorEstimationH\x00\x42\x06\n\x04spec\x1a\xa4\x07\n\x08\x45xternal\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12\x37\n\rop_identifier\x18\x03 \x01(\x0b\x32 .Transform.External.OpIdentifier\x1a\xa1\x04\n\x0cOpIdentifier\x12&\n\x03std\x18\x01 \x01(\x0b\x32\x17.Transform.External.StdH\x00\x12,\n\x06pandas\x18\x02 \x01(\x0b\x32\x1a.Transform.External.PandasH\x00\x12*\n\x05numpy\x18\x03 \x01(\x0b\x32\x19.Transform.External.NumpyH\x00\x12\x34\n\ntensorflow\x18\x04 \x01(\x0b\x32\x1e.Transform.External.TensorflowH\x00\x12.\n\x07sklearn\x18\x05 \x01(\x0b\x32\x1b.Transform.External.SklearnH\x00\x12?\n\x10pandas_profiling\x18\x06 \x01(\x0b\x32#.Transform.External.PandasProfilingH\x00\x12.\n\x07xgboost\x18\x07 \x01(\x0b\x32\x1b.Transform.External.XGBoostH\x00\x12*\n\x05skopt\x18\x08 \x01(\x0b\x32\x19.Transform.External.SkoptH\x00\x12\x30\n\x08imblearn\x18\t \x01(\x0b\x32\x1c.Transform.External.ImblearnH\x00\x12(\n\x04shap\x18\n \x01(\x0b\x32\x18.Transform.External.ShapH\x00\x12*\n\x05scipy\x18\x0b \x01(\x0b\x32\x19.Transform.External.ScipyH\x00\x42\x04\n\x02op\x1a\x13\n\x03Std\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x16\n\x06Pandas\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Numpy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nTensorflow\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07Sklearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1f\n\x0fPandasProfiling\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07XGBoost\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Skopt\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x18\n\x08Imblearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x14\n\x04Shap\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Scipy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\n\n\x08Identity\x1a*\n\x08Variable\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\xa9\x01\n\x08\x43omposed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12@\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\'.Transform.Composed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x34\n\x07Project\x12)\n\nprojection\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x06\x46ilter\x12%\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\t\n\x07Shuffle\x1a)\n\x04Join\x12!\n\x02on\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a+\n\x04\x43\x61st\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\x61\n\x06Sample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1av\n\x0fSchemaInference\x12:\n\x0b\x63\x61st_policy\x18\x01 \x01(\x0e\x32%.Transform.SchemaInference.CastPolicy\"\'\n\nCastPolicy\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bMOST_LIKELY\x10\x01\x1a\x16\n\x07GroupBy\x12\x0b\n\x03key\x18\x01 \x01(\t\x1a\x0b\n\tSynthetic\x1a\x0e\n\x0cUserSettings\x1a/\n\x15\x41utomaticUserSettings\x12\x16\n\x0emax_categories\x18\x01 \x01(\x03\x1a\t\n\x07Protect\x1a\x0b\n\tTranscode\x1a\x12\n\x10InverseTranscode\x1ao\n\x14\x44ifferentiatedSample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1a\x10\n\x0eProtectedPaths\x1a\r\n\x0bPublicPaths\x1a.\n\x07GetItem\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x1a\x0e\n\x0c\x41ssignBudget\x1a\x11\n\x0f\x41utomaticBudget\x1a\x12\n\x10\x41ttributesBudget\x1a\n\n\x08SDBudget\x1a$\n\nDeriveSeed\x12\x16\n\x0erandom_integer\x18\x01 \x01(\x03\x1a\x0b\n\tGroupByPE\x1a\x10\n\x0eSamplingRatios\x1a\x12\n\x10RelationshipSpec\x1a\x88\x01\n\tSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a\x8a\x01\n\x0b\x44PSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a@\n\x0e\x41liasedQueries\x12.\n\raliased_query\x18\x01 \x03(\x0b\x32\x17.Transform.AliasedQuery\x1a\x42\n\x0c\x41liasedQuery\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\r\n\x05query\x18\x02 \x01(\t\x1a\x17\n\x07\x45xtract\x12\x0c\n\x04size\x18\x01 \x01(\x05\x1a\x13\n\x11ValidatedUserType\x1a\x11\n\x0f\x45rrorEstimation\"\x80\x01\n\nSQLDialect\x12\x08\n\x04NONE\x10\x00\x12\x0c\n\x08POSTGRES\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\n\n\x06MY_SQL\x10\x03\x12\x0b\n\x07SQLLITE\x10\x04\x12\n\n\x06ORACLE\x10\x05\x12\r\n\tBIG_QUERY\x10\x06\x12\x0c\n\x08REDSHIFT\x10\x07\x12\x08\n\x04HIVE\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.transform_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSFORM_PROPERTIESENTRY._options = None
  _TRANSFORM_PROPERTIESENTRY._serialized_options = b'8\001'
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._options = None
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _TRANSFORM._serialized_start=158
  _TRANSFORM._serialized_end=4651
  _TRANSFORM_PROPERTIESENTRY._serialized_start=338
  _TRANSFORM_PROPERTIESENTRY._serialized_end=387
  _TRANSFORM_SPEC._serialized_start=390
  _TRANSFORM_SPEC._serialized_end=1961
  _TRANSFORM_EXTERNAL._serialized_start=1964
  _TRANSFORM_EXTERNAL._serialized_end=2896
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_start=2078
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_end=2623
  _TRANSFORM_EXTERNAL_STD._serialized_start=2625
  _TRANSFORM_EXTERNAL_STD._serialized_end=2644
  _TRANSFORM_EXTERNAL_PANDAS._serialized_start=2646
  _TRANSFORM_EXTERNAL_PANDAS._serialized_end=2668
  _TRANSFORM_EXTERNAL_NUMPY._serialized_start=2670
  _TRANSFORM_EXTERNAL_NUMPY._serialized_end=2691
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_start=2693
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_end=2719
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_start=2721
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_end=2744
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_start=2746
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_end=2777
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_start=2779
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_end=2802
  _TRANSFORM_EXTERNAL_SKOPT._serialized_start=2804
  _TRANSFORM_EXTERNAL_SKOPT._serialized_end=2825
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_start=2827
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_end=2851
  _TRANSFORM_EXTERNAL_SHAP._serialized_start=2853
  _TRANSFORM_EXTERNAL_SHAP._serialized_end=2873
  _TRANSFORM_EXTERNAL_SCIPY._serialized_start=2875
  _TRANSFORM_EXTERNAL_SCIPY._serialized_end=2896
  _TRANSFORM_IDENTITY._serialized_start=2898
  _TRANSFORM_IDENTITY._serialized_end=2908
  _TRANSFORM_VARIABLE._serialized_start=2910
  _TRANSFORM_VARIABLE._serialized_end=2952
  _TRANSFORM_COMPOSED._serialized_start=2955
  _TRANSFORM_COMPOSED._serialized_end=3124
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_start=3071
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_end=3124
  _TRANSFORM_PROJECT._serialized_start=3126
  _TRANSFORM_PROJECT._serialized_end=3178
  _TRANSFORM_FILTER._serialized_start=3180
  _TRANSFORM_FILTER._serialized_end=3227
  _TRANSFORM_SHUFFLE._serialized_start=3229
  _TRANSFORM_SHUFFLE._serialized_end=3238
  _TRANSFORM_JOIN._serialized_start=3240
  _TRANSFORM_JOIN._serialized_end=3281
  _TRANSFORM_CAST._serialized_start=3283
  _TRANSFORM_CAST._serialized_end=3326
  _TRANSFORM_SAMPLE._serialized_start=3328
  _TRANSFORM_SAMPLE._serialized_end=3425
  _TRANSFORM_SCHEMAINFERENCE._serialized_start=3427
  _TRANSFORM_SCHEMAINFERENCE._serialized_end=3545
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_start=3506
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_end=3545
  _TRANSFORM_GROUPBY._serialized_start=3547
  _TRANSFORM_GROUPBY._serialized_end=3569
  _TRANSFORM_SYNTHETIC._serialized_start=3571
  _TRANSFORM_SYNTHETIC._serialized_end=3582
  _TRANSFORM_USERSETTINGS._serialized_start=3584
  _TRANSFORM_USERSETTINGS._serialized_end=3598
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_start=3600
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_end=3647
  _TRANSFORM_PROTECT._serialized_start=3649
  _TRANSFORM_PROTECT._serialized_end=3658
  _TRANSFORM_TRANSCODE._serialized_start=3660
  _TRANSFORM_TRANSCODE._serialized_end=3671
  _TRANSFORM_INVERSETRANSCODE._serialized_start=3673
  _TRANSFORM_INVERSETRANSCODE._serialized_end=3691
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_start=3693
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_end=3804
  _TRANSFORM_PROTECTEDPATHS._serialized_start=3806
  _TRANSFORM_PROTECTEDPATHS._serialized_end=3822
  _TRANSFORM_PUBLICPATHS._serialized_start=3824
  _TRANSFORM_PUBLICPATHS._serialized_end=3837
  _TRANSFORM_GETITEM._serialized_start=3839
  _TRANSFORM_GETITEM._serialized_end=3885
  _TRANSFORM_ASSIGNBUDGET._serialized_start=3887
  _TRANSFORM_ASSIGNBUDGET._serialized_end=3901
  _TRANSFORM_AUTOMATICBUDGET._serialized_start=3903
  _TRANSFORM_AUTOMATICBUDGET._serialized_end=3920
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_start=3922
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_end=3940
  _TRANSFORM_SDBUDGET._serialized_start=3942
  _TRANSFORM_SDBUDGET._serialized_end=3952
  _TRANSFORM_DERIVESEED._serialized_start=3954
  _TRANSFORM_DERIVESEED._serialized_end=3990
  _TRANSFORM_GROUPBYPE._serialized_start=3992
  _TRANSFORM_GROUPBYPE._serialized_end=4003
  _TRANSFORM_SAMPLINGRATIOS._serialized_start=4005
  _TRANSFORM_SAMPLINGRATIOS._serialized_end=4021
  _TRANSFORM_RELATIONSHIPSPEC._serialized_start=4023
  _TRANSFORM_RELATIONSHIPSPEC._serialized_end=4041
  _TRANSFORM_SELECTSQL._serialized_start=4044
  _TRANSFORM_SELECTSQL._serialized_end=4180
  _TRANSFORM_DPSELECTSQL._serialized_start=4183
  _TRANSFORM_DPSELECTSQL._serialized_end=4321
  _TRANSFORM_ALIASEDQUERIES._serialized_start=4323
  _TRANSFORM_ALIASEDQUERIES._serialized_end=4387
  _TRANSFORM_ALIASEDQUERY._serialized_start=4389
  _TRANSFORM_ALIASEDQUERY._serialized_end=4455
  _TRANSFORM_EXTRACT._serialized_start=4457
  _TRANSFORM_EXTRACT._serialized_end=4480
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_start=4482
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_end=4501
  _TRANSFORM_ERRORESTIMATION._serialized_start=4503
  _TRANSFORM_ERRORESTIMATION._serialized_end=4520
  _TRANSFORM_SQLDIALECT._serialized_start=4523
  _TRANSFORM_SQLDIALECT._serialized_end=4651
# @@protoc_insertion_point(module_scope)
