syntax = "proto3";
package sarus_data_spec;
import "google/protobuf/any.proto";

message ProtoContainer {
  // General container to store list of protos
  repeated google.protobuf.Any protos = 1; //
  map<string, string> properties = 2; // Other properties
}
