syntax = "proto3";
import "sarus_data_spec/protobuf/statistics.proto";
package sarus_data_spec;

message Size {
  string uuid = 1;
  string dataset = 2; // uuid
  string name = 3;
  Statistics statistics = 4;
  map<string, string> properties = 5;
}
